#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
               ..
              ( '`<
               )(
        ( ----'  '.
        (         ;
         (_______,' 
    ~^~^~^~^~^~^~^~^~^~^~
    Author: thong
    Company: M O B I O
    Date Created: 14/03/2023
"""
from datetime import datetime

from .adapter import DataAdapter


class ElasticsearchAdapter(DataAdapter):

    def execute_query(self, query):
        pass

    def build_query(self, policy):
        """
        :param policy:
        [{
            "effect": "Allow",
            "condition": [{
                "operator": "StringEquals",
                "field": "user:staff_code",
                "values": ["A123456"],
                "qualifier": "ForAnyValue", // ForAllValues
                "if_exists": True,
                "ignore_case": True,
            }]
        }]
        :return:
        """
        must_policy = []
        should_allow = [] # Allow -> or
        must_deny = [] # Deny -> and
        if policy and isinstance(policy, list):
            for pol in policy:
                conditions = pol.get("condition")
                effect = pol.get("effect")
                if not conditions:
                    if effect == "allow":
                        should_allow.append({
                            "match_all": {}
                        })
                    else:
                        must_deny.append({
                            "must_not": [{
                                "match_all": {}
                            }]
                        })
                else:
                    must = []
                    must_not = []
                    should = []
                    for condition in conditions:
                        must_i, must_not_i, should_i = self.convert_operator_to_query(condition)
                        if must_i:
                            must.extend(must_i)
                        if must_not_i:
                            must_not.extend(must_not_i)
                        if should_i:
                            should.extend(should_i)
                    bool_should = {}
                    if must:
                        bool_should["must"] = must
                    if must_not:
                        bool_should["must_not"] = {
                            "bool": {
                                "must": must_not
                            }
                        }
                    if should:
                        bool_should["should"] = should

                    if effect == "allow":
                        should_allow.append({
                            "bool": bool_should
                        })
                    else:
                        must_deny.append({
                            "bool": {
                                "must_not": {
                                    "bool": bool_should
                                }
                            }

                        })
        if should_allow:
            must_policy.append({
                "bool": {
                    "should": should_allow
                }
            })
        if must_deny:
            must_policy.append({
                "bool": {
                    "must": must_deny
                }
            })

        return {
            "must": must_policy,
            "must_not": [],
            "should": []
        }

    def convert_operator_to_query(self, filter):
        must = []
        must_not = []
        should = []

        operator = filter.get("operator")
        field = filter.get("field")
        values = filter.get("values")
        qualifier = filter.get("qualifier")  # default ForAllValues
        date_format = filter.get("date_format")
        if not field or not operator:
            return must, must_not, should

        field_keys = field.split(":")
        field_key = field_keys[1] if len(field_keys) > 1 else None
        if not field_key:
            return must, must_not, should

        # ----- Check String ----
        if operator in ["StringEquals", "NumericEquals"]:
            if qualifier == "ForAnyValue":
                sh = []
                for value in values:
                    if not value:
                        sh.append({
                            "bool": {
                                "must_not": {
                                    "exists": {
                                        "field": field_key
                                    }
                                }
                            }
                        })
                    else:
                        sh.append({"match": {field_key: value}})
                # sh = [{"match": {field_key: value}} if value else {"match": {field_key: ""}} for value in values]
                should.extend(sh)
            else:
                mu = []
                for value in values:
                    if not value:
                        mu.append({
                            "bool": {
                                "must_not": {
                                    "exists": {
                                        "field": field_key
                                    }
                                }
                            }
                        })
                    else:
                        mu.append({"match": {field_key: value}})
                # mu = [{"match": {field_key: value}} if value else {"match": {field_key: ""}} for value in values]
                must.extend(mu)

        elif operator in ["StringNotEquals", "NumericNotEquals"]:
            if qualifier == "ForAnyValue":
                sh = []
                for value in values:
                    if not value:
                        sh.append({
                            "bool": {
                                "must_not": {
                                    "exists": {
                                        "field": field_key
                                    }
                                }
                            }
                        })
                    else:
                        sh.append({"match": {field_key: value}})
                # sh = [{"match": {field_key: value}} if value else {"match": {field_key: ""}} for value in values]
                must_not.append({
                    "bool": {
                        "should": sh
                    }
                })
            else:
                mu = []
                for value in values:
                    if not value:
                        mu.append({
                            "bool": {
                                "must_not": {
                                    "exists": {
                                        "field": field_key
                                    }
                                }
                            }
                        })
                    else:
                        mu.append({"match": {field_key: value}})
                mu = [{"match": {field_key: value}} if value else {"match": {field_key: ""}} for value in values]
                must_not.extend(mu)
        elif operator == "StringContains":
            if qualifier == "ForAnyValue":
                sh = [{"wildcard": {field_key: "*" + value + "*" if value else "*"}} for value in values]
                should.extend(sh)
            else:
                mu = [{"wildcard": {field_key: "*" + value + "*" if value else "*"}} for value in values]
                must.extend(mu)
        elif operator == "StringNotContains":
            if qualifier == "ForAnyValue":
                sh = [{"wildcard": {field_key: "*" + value + "*" if value else "*"}} for value in values]
                must_not.append({
                    "bool": {
                        "should": sh
                    }
                })
            else:
                mu = [{"wildcard": {field_key: "*" + value + "*" if value else "*"}} for value in values]
                must_not.extend(mu)
        elif operator == "StringEndsWith":
            if qualifier == "ForAnyValue":
                sh = [{"wildcard": {field_key: "*" + value if value else "*"}} for value in values]
                should.extend(sh)
            else:
                mu = [{"wildcard": {field_key: "*" + value if value else '*'}} for value in values]
                must.extend(mu)

        elif operator == "StringStartsWith":
            if qualifier == "ForAnyValue":
                sh = [{"wildcard": {field_key: value + "*" if value else "*"}} for value in values]
                should.extend(sh)
            else:
                mu = [{"wildcard": {field_key: value + "*" if value else "*"}} for value in values]
                must.extend(mu)
        elif operator == "StringMatchesRegex":
            if qualifier == "ForAnyValue":
                sh = [{"regexp": {field_key: value if value else ''}} for value in values]
                should.extend(sh)
            else:
                mu = [{"regexp": {field_key: value if value else ''}} for value in values]
                must.extend(mu)
        elif operator == "NumericLessThan":
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lt": value}}} for value in values]
                must.extend(mu)
        elif operator == "NumericLessThanEquals":
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lte": value}}} for value in values]
                must.extend(mu)
        elif operator == "NumericGreaterThan":
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gt": value}}} for value in values]
                must.extend(mu)
        elif operator == "NumericGreaterThanEquals":
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gte": value}}} for value in values]
                must.extend(mu)
        # Tùy data từng bên lưu datetime kiểu gì, Ví dụ Sale lưu theo timestamp thì cần convert ra start_time,
        # end_time trong ngày
        elif operator == "DateEquals":
            for value in values:
                start, end = ElasticsearchAdapter.convert_date_to_range_timestamp(value, date_format)
                if not start or not end:
                    continue
                if qualifier == "ForAnyValue":
                    should.append({"range": {field_key: {"gte": start, "lte": end}}})
                else:
                    must.append({"range": {field_key: {"gte": start, "lte": end}}})
        elif operator == "DateNotEquals":
            sh = []
            for value in values:
                start, end = ElasticsearchAdapter.convert_date_to_range_timestamp(value, date_format)
                if not start or not end:
                    continue
                if qualifier == "ForAnyValue":
                    sh.append({"range": {field_key: {"gte": start, "lte": end}}})
                else:
                    must_not.append({"range": {field_key: {"gte": start, "lte": end}}})
            if sh:
                must_not.append({
                    "bool": {"should": sh}
                })
        elif operator == "DateLessThan":
            values = ElasticsearchAdapter.convert_list_date_to_timestamp(values, date_format)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lt": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateLessThanEquals":
            values = ElasticsearchAdapter.convert_list_date_to_timestamp(values, date_format)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lte": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateGreaterThan":
            values = ElasticsearchAdapter.convert_list_date_to_timestamp(values, date_format)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gt": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateGreaterThanEquals":
            values = ElasticsearchAdapter.convert_list_date_to_timestamp(values, date_format)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gte": value}}} for value in values]
                must.extend(mu)

        # Check Day
        elif operator == "DayEquals":
            for value in values:
                start, end = ElasticsearchAdapter.convert_day_to_range_timestamp(value, date_format)
                if not start or not end:
                    continue
                if qualifier == "ForAnyValue":
                    should.append({"range": {field_key: {"gte": start, "lte": end}}})
                else:
                    must.append({"range": {field_key: {"gte": start, "lte": end}}})
        elif operator == "DayNotEquals":
            sh = []
            for value in values:
                start, end = ElasticsearchAdapter.convert_day_to_range_timestamp(value, date_format)
                if not start or not end:
                    continue
                if qualifier == "ForAnyValue":
                    sh.append({"range": {field_key: {"gte": start, "lte": end}}})
                else:
                    must_not.append({"range": {field_key: {"gte": start, "lte": end}}})
            if sh:
                must_not.append({
                    "bool": {"should": sh}
                })
        elif operator == "DayLessThan":
            values = ElasticsearchAdapter.convert_list_start_day_to_timestamp(values, date_format)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lt": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateLessThanEquals":
            values = ElasticsearchAdapter.convert_list_start_day_to_timestamp(values, date_format)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"lte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"lte": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateGreaterThan":
            values = ElasticsearchAdapter.convert_list_end_day_to_timestamp(values, date_format)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gt": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gt": value}}} for value in values]
                must.extend(mu)
        elif operator == "DateGreaterThanEquals":
            values = ElasticsearchAdapter.convert_list_end_day_to_timestamp(values, date_format)
            if qualifier == "ForAnyValue":
                sh = [{"range": {field_key: {"gte": value}}} for value in values]
                should.extend(sh)
            else:
                mu = [{"range": {field_key: {"gte": value}}} for value in values]
                must.extend(mu)

        elif operator == "ListAllIn":
            for value in values:
                must.append({"term": {field_key: value}})
        elif operator == "ListAllNotIn":
            must_not.append({"terms": {field_key: values}})
        elif operator == "ListAnyIn":
            for value in values:
                should.append({"term": {field_key: value}})
        elif operator == "ListAnyNotIn":
            sh = [{"term": {field_key: value}} for value in values]
            must_not.append({"bool": {"should": sh}})
        elif operator == "ListIsIn":
            must.append({"terms": {field_key: values}})
        elif operator == "ListIsNotIn":
            must_not.append({"terms": {field_key: values}})

        elif operator == "ListIsEmpty":
            must.append({
                "bool": {
                    "must_not": [
                        {
                            "exists": {
                                "field": field_key
                            }
                        }
                    ]
                }
            })
        elif operator == "ListIsNotEmpty":
            must.append({"exists": {"field": field_key}})
        elif operator == "ListAnyContains":
            if qualifier == "ForAnyValue":
                sh = [{"wildcard": {field_key: "*" + value + "*"}} for value in values]
                should.extend(sh)
            else:
                mu = [{"wildcard": {field_key: "*" + value + "*"}} for value in values]
                must.extend(mu)
        elif operator == "ListAnyNotContains":
            if qualifier == "ForAnyValue":
                mu = [{"match": {field_key: value}} if value else {"match": {field_key: ""}} for value in values]
                must_not.extend(mu)
            else:
                sh = [{"match": {field_key: value}} if value else {"match": {field_key: ""}} for value in values]
                must_not.append({
                    "bool": {
                        "should": sh
                    }
                })
        elif operator == "Exists":
            must.append({
                "exists": {
                    "field": field_key
                }
            })
        elif operator == "NotExists":
            must.append({
                "bool": {
                    "should": [
                        {
                            "bool": {
                                "must_not": [
                                    {
                                        "exists": {
                                            "field": field_key
                                        }
                                    }
                                ]
                            }
                        },
                        {
                            "terms": {
                                field_key: []
                            }
                        }
                    ]
                }
            })

        return must, must_not, should

    @staticmethod
    def convert_date_to_range_timestamp(date, date_format):
        """
        :param date:
        :param date_format:
        :return:
        """
        try:
            date_input = datetime.strptime(date, date_format)
            end_time = date_input.replace(second=59).timestamp()
            start_time = end_time - (60)
            return start_time, end_time
        except Exception:
            return None, None

    @staticmethod
    def convert_list_date_to_timestamp(lst_date, date_format):
        """
        :param lst_date:
        :param date_format:
        :return:
        """
        values = []
        for d in lst_date:
            try:
                value = datetime.strptime(d, date_format).timestamp()
                values.append(value)
            except Exception:
                pass
        return values

    @staticmethod
    def convert_day_to_range_timestamp(date, date_format):
        """
        :param date:
        :param date_format:
        :return:
        """
        try:
            date_input = datetime.strptime(date, date_format)
            end_time = date_input.replace(hour=17, minute=0, second=0).timestamp()
            start_time = end_time - (24 * 3600)
            return start_time, end_time
        except Exception:
            return None, None

    @staticmethod
    def convert_list_end_day_to_timestamp(lst_date, date_format):
        """
        :param lst_date:
        :param date_format:
        :return:
        """
        values = []
        for d in lst_date:
            try:
                value = datetime.strptime(d, date_format).replace(hour=17, minute=59, second=59).timestamp()
                values.append(value)
            except Exception:
                pass
        return values

    @staticmethod
    def convert_list_start_day_to_timestamp(lst_date, date_format):
        """
        :param lst_date:
        :return:
        """
        values = []
        for d in lst_date:
            try:
                value = datetime.strptime(d, date_format).replace(hour=17, minute=59, second=59).timestamp() - (
                        24 * 3600)
                values.append(value)
            except Exception:
                pass
        return values



"""
    Any of the values not contains collection conditions
"""

from marshmallow import post_load

from .base import CollectionCondition, CollectionConditionSchema


class AnyNotContains(CollectionCondition):
    """
        Condition for any value of `what` in `values`
    """

    def _is_satisfied(self) -> bool:
        for i in self.values:
            for w in self.what:
                if i in w:
                    return False
        return True


class AnyNotContainsSchema(CollectionConditionSchema):
    """
        JSON schema for any in collection condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return AnyNotContains(**data)



import requests

from .config import lru_cache_redis, Mobio


class APIRequest:
    TimeOut = 10
    ADMIN_GET_FULL_INFO_ACCOUNT = "{domain}/adm/api/v2.1/accounts/{account_id}/full-info"
    ADMIN_GET_LIST_ACTION_FOR_MERCHANT = "{domain}/adm/api/v2.1/policies/actions/sdk"
    ADMIN_GET_LIST_STATEMENT = "{domain}/adm/api/v2.1/statements"


# _TIME_CACHE = 600 if Mobio.VM_TYPE and Mobio.VM_TYPE != "TEST" else 60
_TIME_CACHE = 60


class CallAPI:

    # TODO: thay expiration cho production
    @staticmethod
    @lru_cache_redis.add(expiration=_TIME_CACHE)
    def admin_get_account_info(merchant_id, account_id):
        # try:
        url = APIRequest.ADMIN_GET_FULL_INFO_ACCOUNT.format(domain=Mobio.ADMIN_HOST, account_id=account_id)
        authorization = Mobio.MOBIO_TOKEN
        data_res = requests.get(
            url,
            headers={
                "Authorization": authorization,
                "X-Merchant-ID": merchant_id,
            },
            timeout=APIRequest.TimeOut,
        )
        data = None
        if data_res.status_code == 200:
            if data_res.json().get("data"):
                data = data_res.json().get("data")
        # print("abac_sdk admin_get_account_info: {}".format(data))
        return data
        # except Exception as er:
        #     err_msg = "abac_sdk admin_get_account_info err: {}".format(er)
        #     print(err_msg)
        #     return {}

    @lru_cache_redis.add()
    @staticmethod
    def admin_get_json_action(merchant_id):
        # try:
        url = APIRequest.ADMIN_GET_LIST_ACTION_FOR_MERCHANT.format(domain=Mobio.ADMIN_HOST)
        authorization = Mobio.MOBIO_TOKEN
        data_res = requests.get(
            url,
            headers={
                "Authorization": authorization,
                "X-Merchant-ID": merchant_id,
            },
            timeout=APIRequest.TimeOut,
            params={"page": "-1"}
        )
        data = {}
        if data_res.status_code == 200:
            if data_res.json().get("data"):
                list_data = data_res.json().get("data")
                for i in list_data:
                    data[i.get("resource") + ":" + i.get("key")] = i
        # print("abac_sdk admin_get_json_action: {}".format(data))
        return data
        # except Exception as er:
        #     err_msg = "abac_sdk admin_get_json_action err: {}".format(er)
        #     print(err_msg)
        #     return {}

    # TODO: thay expiration cho production
    @staticmethod
    @lru_cache_redis.add(expiration=_TIME_CACHE)
    def admin_get_list_statement(merchant_id, account_id, resource, action, service):
        """
        :param merchant_id:
        :param account_id:
        :param resource:
        :param action:
        :return: [{
            "effect": "allow",
            "action": [
                "DealViewReport"
            ],
            "resource": [
                "sale:deal"
            ],
            "condition": [
                {
                    "operator": "StringEquals",
                    "field": "user:block",
                    "values": [
                        "KHDN"
                    ],
                    "qualifier": "ForAnyValue",
                    "if_exists": False,
                    "ignore_case": False
                },
                {
                    "operator": "StringEquals",
                    "field": "deal:block",
                    "values": [
                        "${user:block}"
                    ],
                    "qualifier": "ForAnyValue",
                    "if_exists": False,
                    "ignore_case": False
                },
                {
                    "operator": "StringStartsWith",
                    "field": "deal:scope_code",
                    "values": [
                        "${user:scope_code}"
                    ],
                    "qualifier": "ForAnyValue",
                    "if_exists": False,
                    "ignore_case": False
                }
            ]
        }]
        """
        # try:
        url = APIRequest.ADMIN_GET_LIST_STATEMENT.format(domain=Mobio.ADMIN_HOST)
        authorization = Mobio.MOBIO_TOKEN
        body_req = {
            'user_ids': [account_id],
            'team_ids': [],
            'merchant_ids': [merchant_id],
            'resources': service + ":" + resource,
            'actions': resource + ":" + action
        }
        data_res = requests.post(
            url,
            json=body_req,
            headers={
                "Authorization": authorization,
                "X-Merchant-ID": merchant_id,
            },
            timeout=APIRequest.TimeOut,
        )
        data = None
        if data_res.status_code == 200:
            if data_res.json().get("statements"):
                data = data_res.json().get("statements")
        # TODO: comment log khi len prod
        print("abac_sdk admin_get_list_statement: {}".format(data))
        return data

        # return [
        #     {
        #         "effect": "allow",
        #         "action": [
        #             "ListFromSale"
        #         ],
        #         "resource": [
        #             "sale:deal"
        #         ],
        #         "condition": [
        #             {
        #                 "operator": "StringEquals",
        #                 "field": "deal:block",
        #                 "values": [
        #                     "${user:block}"
        #                 ],
        #                 "qualifier": "ForAnyValue",
        #                 "if_exists": False,
        #                 "ignore_case": False
        #             },
        #             {
        #                 "operator": "StringStartsWith",
        #                 "field": "deal:scope_code",
        #                 "values": [
        #                     "${user:scope_code}"
        #                 ],
        #                 "qualifier": "ForAnyValue",
        #                 "if_exists": False,
        #                 "ignore_case": False
        #             }
        #         ]
        #     },
        #             {
        #                 "effect": "allow",
        #                 "action": [
        #                     "ListFromSale"
        #                 ],
        #                 "resource": [
        #                     "sale:deal"
        #                 ],
        #                 "condition": [
        #                     {
        #                         "operator": "StringStartsWith",
        #                         "field": "deal:scope_code",
        #                         "values": [
        #                             "3#7", "3#8", "3#9"
        #                         ],
        #                         "qualifier": "ForAnyValue",
        #                         "if_exists": False,
        #                         "ignore_case": False
        #                     },
        #                     {
        #                         "operator": "StringEquals",
        #                         "field": "user:scope_code",
        #                         "values": [
        #                             "3#6"
        #                         ],
        #                         "qualifier": "ForAnyValue",
        #                         "if_exists": False,
        #                         "ignore_case": False
        #                     }
        #                 ]
        #             }
        # ]

        # except Exception as er:
        #     err_msg = "abac_sdk admin_get_list_statement err: {}".format(er)
        #     print(err_msg)
        #     return []



"""
    Numeric greater than conditions
"""

from marshmallow import post_load

from .base import DateCondition, DateConditionSchema


class DateGt(DateCondition):
    """
        Condition for number `what` greater than `value`
    """

    def _is_satisfied(self) -> bool:
        timestamp_what = self.convert_timestamp_from_any(self.what)
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                timestamp_i = self.convert_timestamp_from_any(i)
                if not timestamp_i or not timestamp_what:
                    return False
                if timestamp_what > timestamp_i:
                    return True
            return False
        else:
            for i in self.values:
                timestamp_i = self.convert_timestamp_from_any(i)
                if not timestamp_i or not timestamp_what:
                    return False
                if timestamp_what <= timestamp_i:
                    return False
            return True


class DateGtSchema(DateConditionSchema):
    """
        JSON schema for greater than datetime condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return DateGt(**data)



"""
    Numeric conditions base class
"""

import logging

from marshmallow import Schema, fields, EXCLUDE, validate

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)


def is_number(value) -> bool:
    """
        Check if value is a number
    """
    return isinstance(value, (float, int))


class NumericCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for numeric conditions
    """

    def __init__(self, values, what, qualifier=ConditionBase.Qualifier.ForAnyValue, **kwargs):
        values_format = [float(i) for i in values]
        self.values = values_format
        self.what = what
        self.qualifier = qualifier


    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is numeric conditions satisfied

            :param what: numeric value to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()


class NumericConditionSchema(Schema):
    """
        Base JSON schema for numeric conditions
    """
    values = fields.List(fields.Number(required=True, allow_none=False), required=True, allow_none=False)
    what = fields.Number(required=True, allow_none=False)
    qualifier = fields.String(allow_none=False, load_default=ConditionBase.Qualifier.ForAnyValue,
                              validate=validate.OneOf(ConditionBase.Qualifier.ALL))
    class Meta:
        unknown = EXCLUDE


"""
    String regex match conditions
"""

import re

from marshmallow import Schema, fields, post_load, ValidationError

from .base import StringCondition


class RegexMatch(StringCondition):
    """
        Condition for string `self.what` matches regex `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if re.search(i, self.what) is not None:
                    return True
            return False
        else:
            for i in self.values:
                if re.search(i, self.what) is None:
                    return False
            return True

def validate_regex(value):
    """
        Validate given regex. Throws ValidationError exception
        for invalid regex expressions.
    """
    # noinspection PyBroadException
    try:
        if isinstance(value, str):
            re.compile(value)
        else:
            for i in value:
                re.compile(i)
    except Exception:
        raise ValidationError("Invalid regex expression '{}'.".format(value))


class RegexMatchSchema(Schema):
    """
        JSON schema for regex match string condition
    """
    values = fields.List(fields.String(required=True, allow_none=False), required=True, allow_none=False, validate=validate_regex)

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return RegexMatch(**data)



from marshmallow import post_load
from .base import DayCondition, DayConditionSchema


class DayGte(DayCondition):

    def _is_satisfied(self) -> bool:
        day_what = self.convert_day_format_from_any(self.what)
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                day_i = self.convert_day_format_from_any(i)
                if day_what >= day_i:
                    return True
            return False
        else:
            for i in self.values:
                day_i = self.convert_day_format_from_any(i)
                if day_what < day_i:
                    return False
            return True


class DayGteSchema(DayConditionSchema):
    """
        JSON schema for greater than datetime condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return DayGte(**data)



from .config import JSONEncoder, KafkaHelper
from .policy.utils import Utils


class ResultAccess:
    def __init__(self):
        self.allow_access = False
        self.display_config = []
        self.filter_config = []
        self.log_statement_deny = []
        self.log_statement_allow = []
        self.log_error = []
        self.data_allow = []
        self.code_error = 403
        self.data_log = {}

    def get_allow_access(self):
        if self.allow_access:
            print("abac_sdk statement_allow: {}".format(self.log_statement_allow))
        else:
            print("abac_sdk statement_deny: {}".format(self.log_statement_deny))
        print("abac_sdk log_error: {}".format(self.log_error))
        self.push_kafka_log()
        return self.allow_access

    def set_allow_access(self, value: bool):
        self.allow_access = value

    def get_filter_config(self):
        """
        filter: [{'effect': 'allow', 'condition': [{'operator': 'StringEquals', 'field': 'deal:block',
        'values': ['KHDN'], 'qualifier': 'ForAnyValue', 'if_exists': False, 'ignore_case': False},
        {'operator': 'StringStartsWith', 'field': 'deal:scope_code', 'values': ['MB##HN'],
        'qualifier': 'ForAnyValue', 'if_exists': False, 'ignore_case': False}]}]
        :return:
        """
        # print(self.filter_config)
        return self.filter_config

    def add_filter_config(self, value):
        value_format = value
        for item in value_format.get("condition", []):
            if item.get("date_format"):
                v_format = []
                for v in item.get("values"):
                    v_format.append(Utils.convert_string_to_format(v, item.get("date_format")))
                item["values"] = v_format
        self.filter_config.append(value_format)

    def get_display_config(self):
        return self.display_config

    def add_display_config(self, value):
        self.display_config.append(value)

    @classmethod
    def build_key_filter_config(cls, effect, operator, field, qualifier):
        return effect + "#" + operator + "#" + field + "#" + qualifier

    @classmethod
    def extract_key_filter_config(cls, key_filter):
        items = key_filter.split("#")
        return {
            "effect": items[0],
            "operator": items[1],
            "field": items[2],
            "qualifier": items[3],
        }

    @classmethod
    def convert_data_save_log(cls, value):
        # return {
        #     'merchant_id': value.get('merchant_id'),
        #     'policy_code': value.get('policy_code'),
        #     'statement_id': value.get('statement_id'),
        #     'effect': value.get('effect'),
        #     'action': value.get('action'),
        #     'resource': value.get('resource'),
        # }
        return JSONEncoder().json_loads(value)

    def add_log_deny(self, value):
        self.log_statement_deny.append(self.convert_data_save_log(value))

    def add_log_allow(self, value):
        self.log_statement_allow.append(self.convert_data_save_log(value))

    def add_log_error(self, value):
        self.log_error.append(value)

    def get_log(self):
        output_engine = {
            "allow_access": self.allow_access,
            "display_config": self.display_config,
            "filter_config": self.filter_config,
            "log_statement_deny": self.log_statement_deny,
            "log_statement_allow": self.log_statement_allow,
            "log_error": self.log_error,
        }
        self.data_log.get("other", {}).update(output_engine)
        self.data_log = JSONEncoder().json_loads(self.data_log)
        return self.data_log

    # gom các condition cùng key ở các statement lại thành 1 condition,
    # nhưng logic chạy list statement thay đổi thành statement or với nhau, các condition trong statement and,
    # như vậy gom sẽ thành sai logic, kể cả gom trong 1 statement cũng là sai logic,
    # ví dụ user nào xem dữ liệu theo mã cấp user đó hoặc 2 mã cấp khác, policy sẽ tạo 2 statment để thỏa mãn logic trên
    def group_condition_key(self):
        """
        filter: [{'effect': 'allow', 'condition': [{'operator': 'StringEquals', 'field': 'deal:block',
        'values': ['KHDN'], 'qualifier': 'ForAnyValue', 'if_exists': False, 'ignore_case': False},
        {'operator': 'StringStartsWith', 'field': 'deal:scope_code', 'values': ['MB##HN'],
        'qualifier': 'ForAnyValue', 'if_exists': False, 'ignore_case': False}]}]
        :return:
        """
        # print(self.filter_config)
        dict_field = {}
        dict_exists = {}
        list_filter_config = []
        index_statement = 0
        for statement in self.filter_config:
            item_condition = []
            index_condition = 0
            effect = statement.get("effect")
            for condition in statement.get("condition"):
                operator = condition.get("operator")
                field = condition.get("field")
                qualifier = condition.get("qualifier")
                key_filter = self.build_key_filter_config(effect, operator, field, qualifier)
                if key_filter not in dict_field:
                    dict_field[key_filter] = [index_statement, index_condition]
                    item_condition.append(condition)
                else:
                    dict_exists[key_filter] = condition
                index_condition += 1
            if item_condition:
                list_filter_config.append({
                    'effect': effect,
                    'condition': item_condition
                })
            index_statement += 1
        for k, v in dict_exists.items():
            if k in dict_field:
                list_filter_config[dict_field.get(k)[0]].get("condition")[dict_field.get(k)[1]].get("values").extend(
                    v.get("values"))
        return list_filter_config

    def get_data_allow(self):
        return self.data_allow

    def add_data_allow(self, value):
        self.data_allow.append(value)

    def get_code_error(self):
        # code 403 ko co quyen do policy merchant tao
        # code 413 ko co quyen do ko co license
        return self.code_error

    def push_kafka_log(self):
        try:
            KafkaHelper.push_message_kafka(topic=KafkaHelper.TOPIC_ABAC_LOG,
                                           data=self.get_log())
        except Exception as err:
            print("push_kafka_log: {}".format(err))



"""
    Collection conditions base class
"""

import logging

from marshmallow import Schema, fields, EXCLUDE

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)


def is_collection(value) -> bool:
    """
        Check if value is a collection
    """
    return any([isinstance(value, list), isinstance(value, set), isinstance(value, tuple)])


class CollectionCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for collection conditions

        :param values: collection of values to compare during policy evaluation
    """

    def __init__(self, values, what, **kwargs):
        self.values = values
        self.what = what


    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is collection conditions satisfied

            :param what: collection to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()


class CollectionConditionSchema(Schema):
    """
        Base JSON schema class for collection conditions
    """
    values = fields.List(
        fields.Raw(required=True, allow_none=False),
        required=True,
        allow_none=False
    )
    what = fields.List(
        fields.Raw(allow_none=False),
        allow_none=False
    )

    class Meta:
        unknown = EXCLUDE


"""
    All of the values not in collection conditions
"""

from marshmallow import post_load

from .base import CollectionCondition, CollectionConditionSchema


class AllNotIn(CollectionCondition):
    """
        Condition for all values of `what` not in `values`
    """

    def _is_satisfied(self) -> bool:
        if not self.what:
            return False
        return not set(self.values).issubset(self.what)


class AllNotInSchema(CollectionConditionSchema):
    """
        JSON schema for all not in collection condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return AllNotIn(**data)



import os
import redis
from mobio.libs.Singleton import Singleton
from mobio.libs.caching import LruCache
# from mobio.libs.logging import MobioLogging
import datetime
import json
import uuid
from bson import ObjectId
import time
from mobio.libs.kafka_lib.helpers.kafka_producer_manager import KafkaProducerManager


class StoreCacheType:
    LOCAL = 1
    REDIS = 2


class Cache:
    REDIS_URI = "{}?health_check_interval=30".format(os.environ.get("ADMIN_REDIS_URI", os.environ.get("REDIS_URI")))
    PREFIX = "m_abac"


lru_cache_redis = LruCache(
    store_type=StoreCacheType.REDIS,
    redis_uri=Cache.REDIS_URI,
    cache_prefix=Cache.PREFIX,
)


# mobio_log = MobioLogging()

class Mobio:
    ADMIN_HOST = os.environ.get("ADMIN_HOST")
    MOBIO_TOKEN = "Basic {}".format(os.environ.get('YEK_REWOP', ''))
    VM_TYPE = os.environ.get("VM_TYPE")

@Singleton
class RedisClient(object):

    def __init__(self):
        self.redis_connect = redis.from_url(Cache.REDIS_URI)

    def get_connect(self):
        return self.redis_connect

    def get_value(self, key_cache):
        redis_conn = self.get_connect()
        return redis_conn.get(key_cache)

    def set_value_expire(self, key_cache, value_cache, time_seconds=3600):
        redis_conn = self.get_connect()
        redis_conn.setex(key_cache, time_seconds, value_cache)

    def delete_key(self, key_cache):
        redis_conn = self.get_connect()
        redis_conn.delete(key_cache)

    class KeyCache:
        MERCHANT_USE_ABAC = "merchant_use_abac"


class JSONEncoder(json.JSONEncoder):
    def __init__(self, format_time="%Y-%m-%dT%H:%M:%S.%fZ"):
        super().__init__()
        self.format_time = format_time

    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)
        elif type(o) == datetime.date:
            a = datetime.datetime.combine(o, datetime.datetime.min.time())
            return a.strftime(self.format_time)
        elif type(o) == datetime.datetime:
            return o.strftime(self.format_time)
        elif isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, (bytes, bytearray)):
            return str(o, "utf-8")
        return json.JSONEncoder.default(self, o)

    def encode_data(self, o):
        if '_id' in o:
            o['_id'] = str(o['_id'])
        return json.JSONEncoder.encode(self, o)

    def json_loads(self, o):
        return json.loads(self.encode_data(o))


class Timer:
    class TimerError(Exception):
        """A custom exception used to report errors in use of Timer class"""

    def __init__(self, title_name=""):
        self._start_time = None
        self.title_name = title_name

    def start(self):
        """Start a new timer"""
        if self._start_time is not None:
            raise Timer.TimerError(f"Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter()

    def stop(self):
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise Timer.TimerError(f"Timer is not running. Use .start() to start it")

        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None
        print(f"{self.title_name} Elapsed time: {elapsed_time:0.4f} seconds")


class KafkaHelper:
    TOPIC_ABAC_LOG = "admin-policy-log"

    @classmethod
    def push_message_kafka(cls, topic: str, data: dict, key=None):
        key = key if key else ""
        KafkaProducerManager().flush_message(topic=topic, key=key, value=data)



"""
    String conditions base class
"""

import logging

from marshmallow import Schema, fields, EXCLUDE, validate

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)


def is_string(value) -> bool:
    """
        Check if value is string
    """
    return isinstance(value, str)


class StringCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for string conditions
    """

    def __init__(self, values, what, qualifier=ConditionBase.Qualifier.ForAnyValue, ignore_case=False,
                 delimiter="", **kwargs):
        self.case_insensitive = ignore_case
        self.values = values
        self.what = what
        self.qualifier = qualifier
        self.delimiter = delimiter


    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is string conditions satisfied

            :param what: string value to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()


class StringConditionSchema(Schema):
    """
        Base JSON schema for string conditions
    """
    values = fields.List(fields.String(required=True, allow_none=False), required=True, allow_none=False)
    ignore_case = fields.Bool(default=False, missing=False)
    what = fields.String(required=True, allow_none=False)
    qualifier = fields.String(allow_none=False, load_default=ConditionBase.Qualifier.ForAnyValue,
                            validate=validate.OneOf(ConditionBase.Qualifier.ALL))
    delimiter = fields.String(required=False, default="", missing="")

    class Meta:
        unknown = EXCLUDE


"""
    String contains conditions
"""

from marshmallow import post_load

from .base import StringCondition, StringConditionSchema


class Contains(StringCondition):
    """
        Condition for string `self.what` contains `value`
        delimiter kiem tra cho field dac biet scope_code
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            # chua bat ky gia tri nao deu se dung
            for i in self.values:
                if self.case_insensitive:
                    if self.delimiter:
                        if i.lower() + self.delimiter in self.what.lower() or i.lower() == self.what.lower():
                            return True
                    else:
                        if i.lower() in self.what.lower():
                            return True
                else:
                    if self.delimiter:
                        if i + self.delimiter in self.what or i == self.what:
                            return True
                    else:
                        if i in self.what:
                            return True
            return False
        else:
            # chua bat ky gia tri nao deu se dung
            for i in self.values:
                if self.case_insensitive:
                    if self.delimiter:
                        if i.lower() + self.delimiter in self.what.lower() or i.lower() == self.what.lower():
                            pass
                        else:
                            return False
                    else:
                        if i.lower() not in self.what.lower():
                            return False
                else:
                    if self.delimiter:
                        if i + self.delimiter in self.what or i == self.what:
                            pass
                        else:
                            return False
                    else:
                        if i not in self.what:
                            return False
            return True

class ContainsSchema(StringConditionSchema):
    """
        JSON schema for contains string conditions
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return Contains(**data)



"""
    String not ends with conditions
"""

from marshmallow import post_load

from .base import StringCondition, StringConditionSchema


class NotEndsWith(StringCondition):
    """
        Condition for string `self.what` ends with `value`
    """

    def _is_satisfied(self,) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if self.case_insensitive:
                    if self.delimiter:
                        if self.what.lower().endswith(self.delimiter + i.lower()) or i.lower() == self.what.lower():
                            return False
                    else:
                        if self.what.lower().endswith(i.lower()):
                            return False
                else:
                    if self.delimiter:
                        if self.what.endswith(self.delimiter + i) or i == self.what:
                            return False
                    else:
                        if self.what.endswith(i):
                            return False
            return True
        else:
            for i in self.values:
                if self.case_insensitive:
                    if self.delimiter:
                        if self.what.lower().endswith(self.delimiter + i.lower()) or i.lower() == self.what.lower():
                            return False
                    else:
                        if self.what.lower().endswith(i.lower()):
                            return False
                else:
                    if self.delimiter:
                        if self.what.endswith(self.delimiter + i) or i == self.what:
                            return False
                    else:
                        if self.what.endswith(i):
                            return False
            return True


class NotEndsWithSchema(StringConditionSchema):
    """
        JSON schema for ends with string condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return NotEndsWith(**data)



"""
    Collection is empty conditions
"""

from marshmallow import Schema, post_load

from .base import CollectionCondition


class IsEmpty(CollectionCondition):
    """
        Condition for `what` being an empty collection
    """


    def _is_satisfied(self) -> bool:
        """
            Is collection conditions satisfied

            :param what: collection to check
            :return: True if satisfied else False
        """
        return len(self.what) == 0


class IsEmptySchema(Schema):
    """
        JSON schema for is empty collection condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use,unused-argument
        # self.validate(data)
        return IsEmpty(**data)



"""
    Any of the values in collection conditions
"""

from marshmallow import post_load

from .base import CollectionCondition, CollectionConditionSchema


class AnyIn(CollectionCondition):
    """
        Condition for any value of `what` in `values`
    """

    def _is_satisfied(self) -> bool:
        if not self.what:
            return False
        return bool(set(self.what).intersection(self.values))


class AnyInSchema(CollectionConditionSchema):
    """
        JSON schema for any in collection condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return AnyIn(**data)



"""
    Operation base class
"""

from abc import ABCMeta, abstractmethod

# from py_abac.context import EvaluationContext
# what is value of field, values is list value policy
MapAnyValue = ""

class ConditionBase(metaclass=ABCMeta):
    """
        Base class for conditions
    """

    # @abstractmethod
    # def is_satisfied(self, ctx: EvaluationContext) -> bool:
    #     """
    #         Is conditions satisfied?
    #
    #         :param ctx: evaluation context
    #         :return: True if satisfied else False
    #     """
    #     raise NotImplementedError()

    class Qualifier:
        ForAnyValue = "ForAnyValue"
        ForAllValues = "ForAllValues"
        ALL = [ForAnyValue, ForAllValues]


"""
    Numeric conditions
"""

from .eq import EqSchema
from .gt import GtSchema
from .gte import GteSchema
from .lt import LtSchema
from .lte import LteSchema
from .neq import NeqSchema
from .list_gt import ListGtSchema
from .list_gte import ListGteSchema
from .list_lt import ListLtSchema
from .list_lte import ListLteSchema



"""
    Numeric greater than conditions
"""

from marshmallow import post_load

from .base_list import NumericListCondition, NumericListConditionSchema


class ListGt(NumericListCondition):
    """
        Condition for number `what` greater than `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for w in self.what:
                for v in self.values:
                    if w > v:
                        return True
            return False
        else:
            for w in self.what:
                for v in self.values:
                    if w <= v:
                        return False
            return True

class ListGtSchema(NumericListConditionSchema):
    """
        JSON schema for greater than numeric condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return ListGt(**data)
